@@ -58,6 +58,10 @@ class Classifier:
         optimize (bool): If True the Boruta algorithm will be run to identify the features
             that contain useful information, after which the optimal Random Forest hyperparameters
             will be calculated using Bayesian optimization. 
+        opt_cv (int): Cross-validations to perform when assesing the performance at each
+            hyperparameter optimization trial. For example, if cv=3, then each optimization trial
+            will be assessed according to the 3-fold cross validation accuracy. Defaults to 10.
+            NOTE: The higher this number, the longer the optimization will take.
         impute (bool): If False no data imputation will be performed. Defaults to True,
             which will result in two outputs, the classifier and the imputer to save
             for future transformations. 
@@ -77,24 +81,29 @@ class Classifier:
         Trained machine learning model.
 
     """
-    def __init__(self, data_x, data_y, clf='rf', optimize=True, impute=True, imp_method='KNN', 
-        n_iter=100, boruta_trials=50, balance=True):
+    def __init__(self, data_x, data_y, clf='rf', optimize=True, opt_cv=10, impute=True, imp_method='KNN', 
+        n_iter=100, boruta_trials=50, boruta_model='rf', balance=True, limit_search=True):
+
         self.data_x = data_x
         self.data_y = data_y
         self.clf = clf
         self.optimize = optimize 
+        self.opt_cv = opt_cv
         self.impute = impute
         self.imp_method = imp_method
         self.n_iter = n_iter
         self.boruta_trials = boruta_trials
+        self.boruta_model = boruta_model
         self.balance = balance 
+        self.limit_search = limit_search
 
         self.model = None
         self.imputer = None
         self.feats_to_use = None
 
-        self.feature_history = None 
+        self.feature_history = None  
         self.optimization_results = None 
+        self.best_params = None 
 
     def create(self):
         """
@@ -131,12 +140,13 @@ class Classifier:
             raise ValueError('clf argument must either be "rf", "nn", or "xgb".')
         
         if self.impute is False and self.optimize is False:
-            print("Returning base model...")
+            print("Returning base {} model...".format(self.clf))
             model.fit(self.data_x, self.data_y)
             self.model = model
             return
 
         if self.impute:
+            print("Imputing data...")
             if self.imp_method == 'KNN':
                 data, self.imputer = KNN_imputation(data=self.data_x, imputer=None)
             elif self.imp_method == 'MissForest':
@@ -145,29 +155,38 @@ class Classifier:
             else:
                 raise ValueError('Invalid imputation method, currently only k-NN and MissForest algorithms are supported.')
             
-            if self.optimize:
-                self.data_x = data
-            else:
+            if self.optimize is False:
+                if data.max() > 1e7 or data.min() < 1e-7:
+                    print('NOTE: Data values higher than 1e7 or lower than 1e-7 will be set to these limits.')
+                    data[data>1e7], data[data<1e-7] = 1e7, 1e-7
+                print("Returning base {} model...".format(self.clf))
                 model.fit(data, self.data_y)
                 self.model = model 
                 return
+                
+        else:
+            data = self.data_x[:]
+            if data.max() > 1e7 or data.min() < 1e-7:
+                print('NOTE: Data values higher than 1e7 or lower than 1e-7 will be set to these limits.')
+                data[data>1e7], data[data<1e-7] = 1e7, 1e-7
 
-        self.feats_to_use, self.feature_history = borutashap_opt(data, self.data_y, boruta_trials=self.boruta_trials)
+        self.feats_to_use, self.feature_history = borutashap_opt(data, self.data_y, boruta_trials=self.boruta_trials, model=self.boruta_model)
         if len(self.feats_to_use) == 0:
-            print('No features selected, increase the number of n_trials when running MicroLIA.optimization.borutashap_opt(). Using all features...')
+            print('No features selected, increase the number of n_trials when running pyBIA.optimization.borutashap_opt(). Using all features...')
             self.feats_to_use = np.arange(data.shape[1])
         #Re-construct the imputer with the selected features as
         #new predictions will only compute these metrics, need to fit again!
         if self.imp_method == 'KNN':
-            self.data_x, self.imputer = KNN_imputation(data=self.data_x[:,self.feats_to_use], imputer=None)
+            data_x, self.imputer = KNN_imputation(data=self.data_x[:,self.feats_to_use], imputer=None)
         elif self.imp_method == 'MissForest':
-            self.data_x, self.imputer = MissForest_imputation(data=self.data_x[:,self.feats_to_use]), None 
+            data_x, self.imputer = MissForest_imputation(data=self.data_x[:,self.feats_to_use]), None 
         else: 
-            self.data_x = self.data_x[:,self.feats_to_use]
+            data_x = self.data_x[:,self.feats_to_use]
 
-        self.model, best_params, self.optimization_results = hyper_opt(self.data_x, self.data_y, clf=self.clf, n_iter=self.n_iter, balance=self.balance, return_study=True)
+        self.model, self.best_params, self.optimization_results = hyper_opt(data_x, self.data_y, clf=self.clf, n_iter=self.n_iter, 
+            balance=self.balance, limit_search=self.limit_search, return_study=True, opt_cv=self.opt_cv)
         print("Fitting and returning final model...")
-        self.model.fit(self.data_x, self.data_y)
+        self.model.fit(data_x, self.data_y)
         
         return
 
@@ -215,6 +234,10 @@ class Classifier:
             joblib.dump(self.feats_to_use, path+'Feats_Index')
         if self.optimization_results is not None:
             joblib.dump(self.optimization_results, path+'HyperOpt_Results')
+        if self.best_params is not None:
+            joblib.dump(self.best_params, path+'Best_Params')
+        if self.feature_history is not None:
+            joblib.dump(self.feature_history, path+'FeatureOpt_Results')
         print('Files saved in: {}'.format(path))
         return 
 
@@ -240,20 +263,27 @@ class Classifier:
         try:
             self.model = joblib.load(path+'Model')
             model = 'model'
-        except FileNotFoundError:
+        except:
             model = ''
             pass
 
         try:
             self.imputer = joblib.load(path+'Imputer')
             imputer = 'imputer'
-        except FileNotFoundError:
+        except:
             imputer = ''
             pass 
 
         try:
             self.feats_to_use = joblib.load(path+'Feats_Index')
             feats_to_use = 'feats_to_use'
+        except:
+            feats_to_use = ''
+            pass
+
+        try:
+            self.feature_history = joblib.load(path+'FeatureOpt_Results')
+            feature_opt_results = 'feature_selection_results'
         except FileNotFoundError:
             feats_to_use = ''
             pass
@@ -261,11 +291,11 @@ class Classifier:
         try:
             self.optimization_results = joblib.load(path+'HyperOpt_Results')
             optimization_results = 'optimization_results'
-        except FileNotFoundError:
+        except:
             optimization_results = '' 
             pass
 
-        print('Successfully loaded the following class attributes: {}, {}, {}, {}'.format(model, imputer, feats_to_use, optimization_results))
+        print('Successfully loaded the following class attributes: {}, {}, {}, {}, {}'.format(model, imputer, feats_to_use, feature_opt_results, optimization_results))
         
         return
 
@@ -312,16 +342,12 @@ class Classifier:
 
         return np.c_[classes, pred[0]]
 
-    def plot_tsne(self, x=None, y=None, norm=False, pca=False, title='Feature Parameter Space'):
+    def plot_tsne(self, norm=False, pca=False, title='Feature Parameter Space',
+        savefig=False):
         """
         Plots a t-SNE projection using the sklearn.manifold.TSNE() method.
 
         Args:
-            x (ndarray, optional): 2D array of size (n x m), where n is the
-                number of samples, and m the number of features. If None then 
-                the data_y attribute will be used. Defaults to None.
-            y (ndarray, optional, str): 1D array containing the corresponing labels.
-                If None then the data_y attribute will be used. Defaults to None.
             norm (bool): If True the data will be min-max normalized. Defaults
                 to False.
             pca (bool): If True the data will be fit to a Principal Component
@@ -331,17 +357,29 @@ class Classifier:
         Returns:
             AxesImage. 
         """
-        if x is None:
-            x = self.data_x 
-        if y is None:
-            y = self.data_y
 
-        if np.any(np.isnan(self.data_x)):
-            print('Automatically imputing NaN values with KNN imputation...')
-            data = KNN_imputation(data=x)[0]
+        if self.model is None:
+            raise ValueError('No model has been created! Run model.create() first.')
+
+        if self.feats_to_use is not None:
+            if len(self.data_x.shape) == 1:
+                data = self.data_x[self.feats_to_use].reshape(1,-1)
+            else:
+                data = self.data_x[:,self.feats_to_use]
         else:
-            data = x
+            data = self.data_x[:]
 
+        if np.any(np.isnan(data)):
+            print('Automatically imputing NaN values with KNN imputation...')
+            if self.imputer is not None and self.imp_method == 'KNN':
+                data = KNN_imputation(data=data, imputer=self.imputer)
+            else:
+                data = KNN_imputation(data=data)[0]
+
+        if data.max() > 1e7 or data.min() < 1e-7:
+            print('NOTE: Data values higher than 1e7 or lower than 1e-7 will be set to these limits.')
+            data[data>1e7], data[data<1e-7] = 1e7, 1e-7
+           
         if len(data) > 5e3:
             method = 'barnes_hut' #Scales with O(N)
         else:
@@ -369,11 +407,18 @@ class Classifier:
             mask = np.where(self.data_y == feat)[0]
             plt.scatter(x[mask], y[mask], marker=markers[count], label=str(feat), alpha=0.7)
 
-        plt.legend(loc='upper right', prop={'size': 16})
+        plt.legend(prop={'size': 16})
         plt.title(title, size=18)
-        plt.show()
+        plt.xticks(fontsize=14)
+        plt.yticks(fontsize=14)
+        if savefig:
+            plt.savefig('tSNE_Projection.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_conf_matrix(self, norm=False, pca=False, k_fold=10, normalize=True, title='Confusion matrix'):
+    def plot_conf_matrix(self, data_y=None, norm=False, pca=False, k_fold=10, normalize=True, title='Confusion matrix',
+        savefig=False):
         """
         Returns a confusion matrix with k-fold validation.
 
@@ -393,20 +438,43 @@ class Classifier:
             normalize (bool, optional): If False the confusion matrix will display the
                 total number of objects in the sample. Defaults to True, in which case
                 the values are normalized between 0 and 1.
-            classes (list): A list containing the label of the two training bags. This
-                will be used to set the axis. Defaults to a list containing 'DIFFUSE' & 'OTHER'. 
+            data_y (list): A list containing the labels of the training bags. This
+                will be used to set the matrix axis. Defaults to None, in which case
+                the data_y will be used. This is useful if you wish to include class
+                labels that are not numerical, since XGB requires numerical labels.
             title (str, optional): The title of the output plot. 
 
         Returns:
             AxesImage.
         """
-        classes = [str(label) for label in np.unique(self.data_y)]
-        if np.any(np.isnan(self.data_x)):
-            print('Automatically imputing NaN values with KNN imputation...')
-            data = KNN_imputation(data=self.data_x)[0]
+        
+        if self.model is None:
+            raise ValueError('No model has been created! Run model.create() first.')
+
+        if data_y is None:
+            classes = [str(label) for label in np.unique(self.data_y)]
         else:
-            data = self.data_x
-     
+            classes = [str(label) for label in np.unique(data_y)]
+
+        if self.feats_to_use is not None:
+            if len(self.data_x.shape) == 1:
+                data = self.data_x[self.feats_to_use].reshape(1,-1)
+            else:
+                data = self.data_x[:,self.feats_to_use]
+        else:
+            data = self.data_x[:]
+
+        if np.any(np.isnan(data)):
+            print('Automatically imputing NaN values with KNN imputation...')
+            if self.imputer is not None and self.imp_method == 'KNN':
+                data = KNN_imputation(data=data, imputer=self.imputer)
+            else:
+                data = KNN_imputation(data=data)[0]
+
+        if data.max() > 1e7 or data.min() < 1e-7:
+            print('NOTE: Data values higher than 1e7 or lower than 1e-7 will be set to these limits.')
+            data[data>1e7], data[data<1e-7] = 1e7, 1e-7
+       
         if norm:
             scaler = MinMaxScaler()
             scaler.fit_transform(data)
@@ -418,9 +486,10 @@ class Classifier:
             data = np.asarray(pca_data).astype('float64')
 
         predicted_target, actual_target = evaluate_model(self.model, data, self.data_y, normalize=normalize, k_fold=k_fold)
-        generate_matrix(predicted_target, actual_target, normalize=normalize, classes=classes, title=title)
+        generate_matrix(predicted_target, actual_target, normalize=normalize, classes=classes, title=title, savefig=savefig)
 
-    def plot_roc_curve(self, k_fold=10, title="Receiver Operating Characteristic Curve"):
+    def plot_roc_curve(self, k_fold=10, title="Receiver Operating Characteristic Curve",
+        savefig=False):
         """
         Plots ROC curve with k-fold cross-validation, as such the 
         standard deviation variations are plotted.
@@ -446,12 +515,28 @@ class Classifier:
         Returns:
             AxesImage
         """
-        if np.any(np.isnan(self.data_x)) == True:
-            print('Automatically imputing NaN values with KNN imputation...')
-            data = KNN_imputation(data=self.data_x)[0]
+        if self.model is None:
+            raise ValueError('No model has been created! Run model.create() first.')
+
+        if self.feats_to_use is not None:
+            if len(self.data_x.shape) == 1:
+                data = self.data_x[self.feats_to_use].reshape(1,-1)
+            else:
+                data = self.data_x[:,self.feats_to_use]
         else:
-            data = self.data_x
-        
+            data = self.data_x[:]
+
+        if np.any(np.isnan(data)):
+            print('Automatically imputing NaN values with KNN imputation...')
+            if self.imputer is not None and self.imp_method == 'KNN':
+                data = KNN_imputation(data=data, imputer=self.imputer)
+            else:
+                data = KNN_imputation(data=data)[0]
+
+        if data.max() > 1e7 or data.min() < 1e-7:
+            print('NOTE: Data values higher than 1e7 or lower than 1e-7 will be set to these limits.')
+            data[data>1e7], data[data<1e-7] = 1e7, 1e-7
+
         model0 = self.model
         if len(np.unique(self.data_y)) != 2:
             X_train, X_test, y_train, y_test = train_test_split(data, self.data_y, test_size=0.2, random_state=0)
@@ -521,13 +606,20 @@ class Classifier:
         plt.ylabel('True Positive Rate', size=14)
         plt.xlabel('False Positive Rate', size=14)
         plt.title(label=title,fontsize=18)
-        plt.show()
+        if savefig:
+            plt.savefig('Ensemble_ROC_Curve.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_hyper_opt(xlog=True, ylog=False):
+    def plot_hyper_opt(self, xlim=None, ylim=None, xlog=True, ylog=False, 
+        savefig=False):
         """
         Plots the hyperparameter optimization history.
     
         Args:
+            xlim: Limits for the x-axis. Ex) xlim = (0, 1000)
+            ylim: Limits for the y-axis. Ex) ylim = (0.9, 0.94)
             xlog (boolean): If True the x-axis will be log-scaled.
                 Defaults to True.
             ylog (boolean): If True the y-axis will be log-scaled.
@@ -537,23 +629,30 @@ class Classifier:
             AxesImage
         """
 
-        fig = optuna.visualization.matplotlib.plot_optimization_history(self.optimization_results)
+        fig = plot_optimization_history(self.optimization_results)
         if xlog:
             plt.xscale('log')
         if ylog:
             plt.yscale('log')
         plt.xlabel('Trial #', size=16)
-        plt.ylabel('10-fold CV Accuracy', size=16)
+        plt.ylabel('10-Fold CV Accuracy', size=16)
         plt.title(('Hyperparameter Optimization History'), size=18)
-        plt.xlim((1,1e4))
-        plt.ylim((0.9, 0.935))
         plt.xticks(fontsize=14)
         plt.yticks(fontsize=14)
         plt.grid(True, color='k', alpha=0.35, linewidth=1.5, linestyle='--')
-        plt.legend(prop={'size': 16})
-        plt.show()
+        plt.legend(prop={'size': 16}, loc=2)
+        if xlim is not None:
+            plt.xlim(xlim)
+        if ylim is not None:
+            plt.ylim(ylim)
+        plt.tight_layout()
+        if savefig:
+            plt.savefig('Ensemble_Hyperparameter_Optimization.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_feature_opt(feats='all'):
+    def plot_feature_opt(self, feats='all'):
         """
         Returns whisker plot displaying the z-score distribution of each feature
         across all trials.
@@ -566,8 +665,7 @@ class Classifier:
             AxesImage
         """
 
-        self.feat_selector.plot(which_features=feats, X_size=14)
-
+        self.feature_history.plot(which_features=feats, X_size=14)
 
 #Helper functions below to generate confusion matrix
 def evaluate_model(classifier, data_x, data_y, normalize=True, k_fold=10):
@@ -605,7 +703,8 @@ def evaluate_model(classifier, data_x, data_y, normalize=True, k_fold=10):
 
     return predicted_targets, actual_targets
 
-def generate_matrix(predicted_labels_list, actual_targets, normalize=True, classes=["DIFFUSE","OTHER"], title='Confusion matrix'):
+def generate_matrix(predicted_labels_list, actual_targets, classes, normalize=True, title='Confusion matrix',
+    savefig=False):
     """
     Generates the confusion matrix using the output from the evaluate_model() function.
 
@@ -630,7 +729,12 @@ def generate_matrix(predicted_labels_list, actual_targets, normalize=True, class
         generate_plot(conf_matrix, classes=classes, normalize=normalize, title=title)
     elif normalize == False:
         generate_plot(conf_matrix, classes=classes, normalize=normalize, title=title)
-    plt.show()
+
+    if savefig:
+        plt.savefig('Ensemble_Confusion_Matrix.png', bbox_inches='tight', dpi=300)
+        plt.clf()
+    else:
+        plt.show()
 
 def generate_plot(conf_matrix, classes, normalize=False, title='Confusion Matrix'):
     """
